@@ -2,7 +2,7 @@
 
   ruby.c -
 
-  $Author: nagachika $
+  $Author: nobu $
   created at: Tue Aug 10 12:47:31 JST 1993
 
   Copyright (C) 1993-2007 Yukihiro Matsumoto
@@ -122,15 +122,6 @@ enum dump_flag_bits {
 typedef struct ruby_cmdline_options ruby_cmdline_options_t;
 
 struct ruby_cmdline_options {
-    int sflag, xflag;
-    int do_loop, do_print;
-    int do_line, do_split;
-    int do_search;
-    unsigned int features;
-    int verbose;
-    int safe_level;
-    unsigned int setids;
-    unsigned int dump;
     const char *script;
     VALUE script_name;
     VALUE e_script;
@@ -141,7 +132,18 @@ struct ruby_cmdline_options {
 	} enc;
     } src, ext, intern;
     VALUE req_list;
+    unsigned int features;
+    unsigned int dump;
+    int safe_level;
+    int sflag, xflag;
     unsigned int warning: 1;
+    unsigned int verbose: 1;
+    unsigned int do_loop: 1;
+    unsigned int do_print: 1;
+    unsigned int do_line: 1;
+    unsigned int do_split: 1;
+    unsigned int do_search: 1;
+    unsigned int setids: 2;
 };
 
 static void init_ids(ruby_cmdline_options_t *);
@@ -175,8 +177,10 @@ cmdline_options_init(ruby_cmdline_options_t *opt)
     return opt;
 }
 
-static NODE *load_file(VALUE, VALUE, int, ruby_cmdline_options_t *);
-static void forbid_setid(const char *, ruby_cmdline_options_t *);
+static rb_ast_t *load_file(VALUE parser, VALUE fname, VALUE f, int script,
+		       ruby_cmdline_options_t *opt);
+static VALUE open_load_file(VALUE fname_v, int *xflag);
+static void forbid_setid(const char *, const ruby_cmdline_options_t *);
 #define forbid_setid(s) forbid_setid((s), opt)
 
 static struct {
@@ -236,13 +240,21 @@ usage(const char *name, int help)
 	M("-h",		   "",			   "show this message, --help for more info"),
     };
     static const struct message help_msg[] = {
-	M("--copyright",                   "", "print the copyright"),
-	M("--enable=feature[,...]",	   ", --disable=feature[,...]",
-	  "enable or disable features"),
-	M("--external-encoding=encoding",  ", --internal-encoding=encoding",
+	M("--copyright",                            "", "print the copyright"),
+	M("--dump={insns|parsetree|...}[,...]",     "",
+          "dump debug information. see below for available dump list"),
+	M("--enable={gems|rubyopt|...}[,...]", ", --disable={gems|rubyopt|...}[,...]",
+	  "enable or disable features. see below for available features"),
+	M("--external-encoding=encoding",           ", --internal-encoding=encoding",
 	  "specify the default external or internal character encoding"),
-	M("--version",                     "", "print the version"),
-	M("--help",			   "", "show this message, -h for short message"),
+	M("--version",                              "", "print the version"),
+	M("--help",			            "", "show this message, -h for short message"),
+    };
+    static const struct message dumps[] = {
+	M("insns",                  "", "instruction sequences"),
+	M("yydebug",                "", "yydebug of yacc parser generator"),
+	M("parsetree",              "", "AST"),
+	M("parsetree_with_comment", "", "AST with comments"),
     };
     static const struct message features[] = {
 	M("gems",    "",        "rubygems (default: "DEFAULT_RUBYGEMS_ENABLED")"),
@@ -262,6 +274,9 @@ usage(const char *name, int help)
 
     for (i = 0; i < numberof(help_msg); ++i)
 	SHOW(help_msg[i]);
+    puts("Dump List:");
+    for (i = 0; i < numberof(dumps); ++i)
+	SHOW(dumps[i]);
     puts("Features:");
     for (i = 0; i < numberof(features); ++i)
 	SHOW(features[i]);
@@ -423,6 +438,8 @@ str_conv_enc(VALUE str, rb_encoding *from, rb_encoding *to)
 				ECONV_UNDEF_REPLACE|ECONV_INVALID_REPLACE,
 				Qnil);
 }
+#else
+# define str_conv_enc(str, from, to) (str)
 #endif
 
 void ruby_init_loadpath_safe(int safe_level);
@@ -433,7 +450,7 @@ ruby_init_loadpath(void)
     ruby_init_loadpath_safe(0);
 }
 
-#if defined(LOAD_RELATIVE) && defined(HAVE_DLADDR)
+#if defined(LOAD_RELATIVE) && defined(HAVE_DLADDR) && !defined(__CYGWIN__)
 static VALUE
 dladdr_path(const void* addr)
 {
@@ -444,7 +461,7 @@ dladdr_path(const void* addr)
 	return rb_str_new(0, 0);
     }
 #ifdef __linux__
-    else if (dli.dli_fname == origarg.argv[0]) {
+    else if (origarg.argc > 0 && origarg.argv && dli.dli_fname == origarg.argv[0]) {
 	fname = rb_str_new_cstr("/proc/self/exe");
 	path = rb_readlink(fname, NULL);
     }
@@ -725,11 +742,11 @@ moreswitches(const char *s, ruby_cmdline_options_t *opt, int envopt)
 
     while (ISSPACE(*s)) s++;
     if (!*s) return;
-    argstr = rb_str_tmp_new((len = strlen(s)) + 2);
+    argstr = rb_str_tmp_new((len = strlen(s)) + (envopt!=0));
     argary = rb_str_tmp_new(0);
 
     p = RSTRING_PTR(argstr);
-    *p++ = ' ';
+    if (envopt) *p++ = ' ';
     memcpy(p, s, len + 1);
     ap = 0;
     rb_str_cat(argary, (char *)&ap, sizeof(ap));
@@ -746,7 +763,7 @@ moreswitches(const char *s, ruby_cmdline_options_t *opt, int envopt)
     rb_str_cat(argary, (char *)&ap, sizeof(ap));
     argv = (char **)RSTRING_PTR(argary);
 
-    while ((i = proc_options(argc, argv, opt, envopt)) > 1 && (argc -= i) > 0) {
+    while ((i = proc_options(argc, argv, opt, envopt)) > 1 && envopt && (argc -= i) > 0) {
 	argv += i;
 	if (**argv != '-') {
 	    *--*argv = '-';
@@ -850,12 +867,18 @@ disable_option(const char *str, int len, void *arg)
     feature_option(str, len, arg, 0U);
 }
 
+RUBY_EXTERN const int  ruby_patchlevel;
+int ruby_env_debug_option(const char *str, int len, void *arg);
+
 static void
 debug_option(const char *str, int len, void *arg)
 {
     static const char list[] = EACH_DEBUG_FEATURES(LITERAL_NAME_ELEMENT, ", ");
 #define SET_WHEN_DEBUG(bit) SET_WHEN(#bit, DEBUG_BIT(bit), str, len)
     EACH_DEBUG_FEATURES(SET_WHEN_DEBUG, ;);
+#ifdef RUBY_DEVEL
+    if (ruby_patchlevel < 0 && ruby_env_debug_option(str, len, 0)) return;
+#endif
     rb_warn("unknown argument for --debug: `%.*s'", len, str);
     rb_warn("debug features are [%.*s].", (int)strlen(list), list);
 }
@@ -900,7 +923,7 @@ proc_options(long argc, char **argv, ruby_cmdline_options_t *opt, int envopt)
     const char *s;
     int warning = opt->warning;
 
-    if (argc == 0)
+    if (argc <= 0 || !argv)
 	return 0;
 
     for (argc--, argv++; argc > 0; argc--, argv++) {
@@ -1050,6 +1073,7 @@ proc_options(long argc, char **argv, ruby_cmdline_options_t *opt, int envopt)
 
 	  case 'x':
 	    if (envopt) goto noenvopt;
+	    forbid_setid("-x");
 	    opt->xflag = TRUE;
 	    s++;
 	    if (*s && chdir(s) < 0) {
@@ -1437,8 +1461,9 @@ rb_f_chomp(int argc, VALUE *argv)
 static VALUE
 process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
 {
-    NODE *tree = 0;
+    rb_ast_t *ast = 0;
     VALUE parser;
+    VALUE script_name;
     const rb_iseq_t *iseq;
     rb_encoding *enc, *lenc;
 #if UTF8_PATH
@@ -1451,14 +1476,18 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
     const struct rb_block *base_block;
     unsigned int dump = opt->dump & dump_exit_bits;
 
-    argc -= i;
-    argv += i;
-
     if (opt->dump & (DUMP_BIT(usage)|DUMP_BIT(help))) {
-	usage(origarg.argv[0], (opt->dump & DUMP_BIT(help)));
+	const char *const progname =
+	    (argc > 0 && argv && argv[0] ? argv[0] :
+	     origarg.argc > 0 && origarg.argv && origarg.argv[0] ? origarg.argv[0] :
+	     ruby_engine);
+	usage(progname, (opt->dump & DUMP_BIT(help)));
 	return Qtrue;
     }
 
+    argc -= i;
+    argv += i;
+
     if ((opt->features & FEATURE_BIT(rubyopt)) &&
 	opt->safe_level == 0 && (s = getenv("RUBYOPT"))) {
 	VALUE src_enc_name = opt->src.enc.name;
@@ -1488,7 +1517,7 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
     }
 
     if (!opt->e_script) {
-	if (argc == 0) {	/* no more args */
+	if (argc <= 0) {	/* no more args */
 	    if (opt->verbose)
 		return Qtrue;
 	    opt->script = "-";
@@ -1514,6 +1543,9 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
 	    argc--;
 	    argv++;
 	}
+	if (opt->script[0] == '-' && !opt->script[1]) {
+	    forbid_setid("program input from stdin");
+	}
     }
 
     opt->script_name = rb_str_new_cstr(opt->script);
@@ -1560,9 +1592,17 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
 	ienc = enc;
 #endif
     }
-    rb_enc_associate(opt->script_name, lenc);
+    script_name = opt->script_name;
+    rb_enc_associate(opt->script_name,
+		     IF_UTF8_PATH(uenc = rb_utf8_encoding(), lenc));
+#if UTF8_PATH
+    if (uenc != lenc) {
+	opt->script_name = str_conv_enc(opt->script_name, uenc, lenc);
+	opt->script = RSTRING_PTR(opt->script_name);
+    }
+#endif
     rb_obj_freeze(opt->script_name);
-    if (IF_UTF8_PATH((uenc = rb_utf8_encoding()) != lenc, 1)) {
+    if (IF_UTF8_PATH(uenc != lenc, 1)) {
 	long i;
 	VALUE load_path = GET_VM()->load_path;
 	const ID id_initial_load_path_mark = INITIAL_LOAD_PATH_MARK;
@@ -1598,12 +1638,6 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
 	rb_funcallv(rb_cISeq, rb_intern_const("compile_option="), 1, &option);
 #undef SET_COMPILE_OPTION
     }
-#if UTF8_PATH
-    if (uenc != lenc) {
-	opt->script_name = str_conv_enc(opt->script_name, uenc, lenc);
-	opt->script = RSTRING_PTR(opt->script_name);
-    }
-#endif
     ruby_set_argv(argc, argv);
     process_sflag(&opt->sflag);
 
@@ -1612,6 +1646,9 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
     /* need to acquire env from toplevel_binding each time, since it
      * may update after eval() */
 
+    base_block = toplevel_context(toplevel_binding);
+    rb_parser_set_context(parser, base_block, TRUE);
+
     if (opt->e_script) {
 	VALUE progname = rb_progname;
 	rb_encoding *eenc;
@@ -1635,19 +1672,14 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
 	    require_libraries(&opt->req_list);
 	}
         ruby_set_script_name(progname);
-
-	base_block = toplevel_context(toplevel_binding);
-	rb_parser_set_context(parser, base_block, TRUE);
-	tree = rb_parser_compile_string(parser, opt->script, opt->e_script, 1);
+	rb_parser_set_options(parser, opt->do_print, opt->do_loop,
+			      opt->do_line, opt->do_split);
+	ast = rb_parser_compile_string(parser, opt->script, opt->e_script, 1);
     }
     else {
-	if (opt->script[0] == '-' && !opt->script[1]) {
-	    forbid_setid("program input from stdin");
-	}
-
-	base_block = toplevel_context(toplevel_binding);
-	rb_parser_set_context(parser, base_block, TRUE);
-	tree = load_file(parser, opt->script_name, 1, opt);
+	VALUE f;
+	f = open_load_file(script_name, &opt->xflag);
+	ast = load_file(parser, opt->script_name, f, 1, opt);
     }
     ruby_set_script_name(opt->script_name);
     if (dump & DUMP_BIT(yydebug)) {
@@ -1672,7 +1704,10 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
 	rb_enc_set_default_internal(Qnil);
     rb_stdio_set_default_encoding();
 
-    if (!tree) return Qfalse;
+    if (!ast->root) {
+	rb_ast_dispose(ast);
+	return Qfalse;
+    }
 
     process_sflag(&opt->sflag);
     opt->xflag = 0;
@@ -1683,11 +1718,7 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
 	if (!dump) return Qtrue;
     }
 
-    if (opt->do_print) {
-	tree = rb_parser_append_print(parser, tree);
-    }
     if (opt->do_loop) {
-	tree = rb_parser_while_loop(parser, tree, opt->do_line, opt->do_split);
 	rb_define_global_function("sub", rb_f_sub, -1);
 	rb_define_global_function("gsub", rb_f_gsub, -1);
 	rb_define_global_function("chop", rb_f_chop, 0);
@@ -1695,19 +1726,31 @@ process_options(int argc, char **argv, ruby_cmdline_options_t *opt)
     }
 
     if (dump & (DUMP_BIT(parsetree)|DUMP_BIT(parsetree_with_comment))) {
-	rb_io_write(rb_stdout, rb_parser_dump_tree(tree, dump & DUMP_BIT(parsetree_with_comment)));
+	rb_io_write(rb_stdout, rb_parser_dump_tree(ast->root, dump & DUMP_BIT(parsetree_with_comment)));
 	rb_io_flush(rb_stdout);
 	dump &= ~DUMP_BIT(parsetree)&~DUMP_BIT(parsetree_with_comment);
-	if (!dump) return Qtrue;
+	if (!dump) {
+	    rb_ast_dispose(ast);
+	    return Qtrue;
+	}
     }
 
     {
 	VALUE path = Qnil;
 	if (!opt->e_script && strcmp(opt->script, "-")) {
-	    path = rb_realpath_internal(Qnil, opt->script_name, 1);
+	    path = rb_realpath_internal(Qnil, script_name, 1);
+#if UTF8_PATH
+	    if (uenc != lenc) {
+		path = str_conv_enc(path, uenc, lenc);
+	    }
+#endif
+	    if (!ENCODING_GET(path)) { /* ASCII-8BIT */
+		rb_enc_copy(path, opt->script_name);
+	    }
 	}
 	base_block = toplevel_context(toplevel_binding);
-	iseq = rb_iseq_new_main(tree, opt->script_name, path, vm_block_iseq(base_block));
+	iseq = rb_iseq_new_main(ast->root, opt->script_name, path, vm_block_iseq(base_block));
+	rb_ast_dispose(ast);
     }
 
     if (dump & DUMP_BIT(insns)) {
@@ -1732,7 +1775,7 @@ static void
 warn_cr_in_shebang(const char *str, long len)
 {
     if (str[len-1] == '\n' && str[len-2] == '\r') {
-	rb_warn("shebang line ends with \\r may cause a problem");
+	rb_warn("shebang line ending with \\r may cause problems");
     }
 }
 #else
@@ -1743,7 +1786,6 @@ struct load_file_arg {
     VALUE parser;
     VALUE fname;
     int script;
-    int xflag;
     ruby_cmdline_options_t *opt;
     VALUE f;
 };
@@ -1758,12 +1800,10 @@ load_file_internal(VALUE argp_v)
     ruby_cmdline_options_t *opt = argp->opt;
     VALUE f = argp->f;
     int line_start = 1;
-    NODE *tree = 0;
+    rb_ast_t *ast = 0;
     rb_encoding *enc;
     ID set_encoding;
-    int xflag = argp->xflag;
 
-    argp->script = 0;
     CONST_ID(set_encoding, "set_encoding");
     if (script) {
 	VALUE c = 1;		/* something not nil */
@@ -1777,11 +1817,9 @@ load_file_internal(VALUE argp_v)
 	enc = rb_ascii8bit_encoding();
 	rb_funcall(f, set_encoding, 1, rb_enc_from_encoding(enc));
 
-	if (xflag || opt->xflag) {
+	if (opt->xflag) {
 	    line_start--;
 	  search_shebang:
-	    forbid_setid("-x");
-	    opt->xflag = FALSE;
 	    while (!NIL_P(line = rb_io_gets(f))) {
 		line_start++;
 		RSTRING_GETMEM(line, str, len);
@@ -1842,8 +1880,7 @@ load_file_internal(VALUE argp_v)
 	    rb_io_ungetbyte(f, c);
 	}
 	else {
-	    if (f != rb_stdin) rb_io_close(f);
-	    f = Qnil;
+	    argp->f = f = Qnil;
 	}
 	if (!(opt->dump & ~DUMP_BIT(version_v))) {
 	    ruby_set_script_name(opt->script_name);
@@ -1859,22 +1896,40 @@ load_file_internal(VALUE argp_v)
     else {
 	enc = rb_utf8_encoding();
     }
+    rb_parser_set_options(parser, opt->do_print, opt->do_loop,
+			  opt->do_line, opt->do_split);
     if (NIL_P(f)) {
 	f = rb_str_new(0, 0);
 	rb_enc_associate(f, enc);
 	return (VALUE)rb_parser_compile_string_path(parser, orig_fname, f, line_start);
     }
     rb_funcall(f, set_encoding, 2, rb_enc_from_encoding(enc), rb_str_new_cstr("-"));
-    tree = rb_parser_compile_file_path(parser, orig_fname, f, line_start);
+    ast = rb_parser_compile_file_path(parser, orig_fname, f, line_start);
     rb_funcall(f, set_encoding, 1, rb_parser_encoding(parser));
-    if (script && rb_parser_end_seen_p(parser)) argp->script = script;
-    return (VALUE)tree;
+    if (script && rb_parser_end_seen_p(parser)) {
+	/*
+	 * DATA is a File that contains the data section of the executed file.
+	 * To create a data section use <tt>__END__</tt>:
+	 *
+	 *   $ cat t.rb
+	 *   puts DATA.gets
+	 *   __END__
+	 *   hello world!
+	 *
+	 *   $ ruby t.rb
+	 *   hello world!
+	 */
+	rb_define_global_const("DATA", f);
+	argp->f = Qnil;
+    }
+    return (VALUE)ast;
 }
 
 static VALUE
 open_load_file(VALUE fname_v, int *xflag)
 {
-    const char *fname = StringValueCStr(fname_v);
+    const char *fname = (fname_v = rb_str_encode_ospath(fname_v),
+			 StringValueCStr(fname_v));
     long flen = RSTRING_LEN(fname_v);
     VALUE f;
     int e;
@@ -1953,39 +2008,23 @@ restore_load_file(VALUE arg)
     struct load_file_arg *argp = (struct load_file_arg *)arg;
     VALUE f = argp->f;
 
-    if (argp->script) {
-	/*
-	 * DATA is a File that contains the data section of the executed file.
-	 * To create a data section use <tt>__END__</tt>:
-	 *
-	 *   $ cat t.rb
-	 *   puts DATA.gets
-	 *   __END__
-	 *   hello world!
-	 *
-	 *   $ ruby t.rb
-	 *   hello world!
-	 */
-	rb_define_global_const("DATA", f);
-    }
-    else if (f != rb_stdin) {
+    if (!NIL_P(f) && f != rb_stdin) {
 	rb_io_close(f);
     }
     return Qnil;
 }
 
-static NODE *
-load_file(VALUE parser, VALUE fname, int script, ruby_cmdline_options_t *opt)
+static rb_ast_t *
+load_file(VALUE parser, VALUE fname, VALUE f, int script, ruby_cmdline_options_t *opt)
 {
     struct load_file_arg arg;
     arg.parser = parser;
     arg.fname = fname;
     arg.script = script;
     arg.opt = opt;
-    arg.xflag = 0;
-    arg.f = open_load_file(rb_str_encode_ospath(fname), &arg.xflag);
-    return (NODE *)rb_ensure(load_file_internal, (VALUE)&arg,
-			     restore_load_file, (VALUE)&arg);
+    arg.f = f;
+    return (rb_ast_t *)rb_ensure(load_file_internal, (VALUE)&arg,
+			      restore_load_file, (VALUE)&arg);
 }
 
 void *
@@ -1998,17 +2037,15 @@ rb_load_file(const char *fname)
 void *
 rb_load_file_str(VALUE fname_v)
 {
-    ruby_cmdline_options_t opt;
-
-    return load_file(rb_parser_new(), fname_v, 0, cmdline_options_init(&opt));
+    return rb_parser_load_file(rb_parser_new(), fname_v);
 }
 
 void *
 rb_parser_load_file(VALUE parser, VALUE fname_v)
 {
     ruby_cmdline_options_t opt;
-
-    return load_file(parser, fname_v, 0, cmdline_options_init(&opt));
+    VALUE f = open_load_file(fname_v, &cmdline_options_init(&opt)->xflag);
+    return load_file(parser, fname_v, f, 0, &opt);
 }
 
 /*
@@ -2028,6 +2065,8 @@ proc_argv0(VALUE process)
     return rb_orig_progname;
 }
 
+static VALUE ruby_setproctitle(VALUE title);
+
 /*
  *  call-seq:
  *     Process.setproctitle(string)  -> string
@@ -2048,10 +2087,14 @@ proc_argv0(VALUE process)
 static VALUE
 proc_setproctitle(VALUE process, VALUE title)
 {
-    StringValue(title);
-
-    setproctitle("%.*s", RSTRING_LENINT(title), RSTRING_PTR(title));
+    return ruby_setproctitle(title);
+}
 
+static VALUE
+ruby_setproctitle(VALUE title)
+{
+    const char *ptr = StringValueCStr(title);
+    setproctitle("%.*s", RSTRING_LENINT(title), ptr);
     return title;
 }
 
@@ -2061,7 +2104,7 @@ set_arg0(VALUE val, ID id)
     if (origarg.argv == 0)
 	rb_raise(rb_eRuntimeError, "$0 not initialized");
 
-    rb_progname = rb_str_new_frozen(proc_setproctitle(rb_mProcess, val));
+    rb_progname = rb_str_new_frozen(ruby_setproctitle(val));
 }
 
 static inline VALUE
@@ -2117,7 +2160,7 @@ init_ids(ruby_cmdline_options_t *opt)
 
 #undef forbid_setid
 static void
-forbid_setid(const char *s, ruby_cmdline_options_t *opt)
+forbid_setid(const char *s, const ruby_cmdline_options_t *opt)
 {
     if (opt->setids & 1)
         rb_raise(rb_eSecurityError, "no %s allowed while running setuid", s);
@@ -2183,9 +2226,9 @@ ruby_set_argv(int argc, char **argv)
     VALUE av = rb_argv;
 
 #if defined(USE_DLN_A_OUT)
-    if (origarg.argv)
+    if (origarg.argc > 0 && origarg.argv)
 	dln_argv0 = origarg.argv[0];
-    else
+    else if (argc > 0 && argv)
 	dln_argv0 = argv[0];
 #endif
     rb_ary_clear(av);
@@ -2204,6 +2247,10 @@ ruby_process_options(int argc, char **argv)
     VALUE iseq;
     const char *script_name = (argc > 0 && argv[0]) ? argv[0] : ruby_engine;
 
+    if (!origarg.argv || origarg.argc <= 0) {
+	origarg.argc = argc;
+	origarg.argv = argv;
+    }
     ruby_script(script_name);  /* for the time being */
     rb_argv0 = rb_str_new4(rb_progname);
     rb_gc_register_mark_object(rb_argv0);
@@ -2246,11 +2293,12 @@ fill_standard_fds(void)
     }
 }
 
-/*! Initializes the process for ruby(1).
+/*! Initializes the process for libruby.
  *
  * This function assumes this process is ruby(1) and it has just started.
- * Usually programs that embeds CRuby interpreter should not call this function,
- * and should do their own initialization.
+ * Usually programs that embed CRuby interpreter may not call this function,
+ * and may do their own initialization.
+ * argc and argv cannot be NULL.
  */
 void
 ruby_sysinit(int *argc, char ***argv)
@@ -2258,10 +2306,12 @@ ruby_sysinit(int *argc, char ***argv)
 #if defined(_WIN32)
     rb_w32_sysinit(argc, argv);
 #endif
-    origarg.argc = *argc;
-    origarg.argv = *argv;
+    if (*argc >= 0 && *argv) {
+	origarg.argc = *argc;
+	origarg.argv = *argv;
 #if defined(USE_DLN_A_OUT)
-    dln_argv0 = origarg.argv[0];
+	dln_argv0 = origarg.argv[0];
 #endif
+    }
     fill_standard_fds();
 }
